{=========================================================================
  RkLaunch Unit

  This unit implements the TRkLauncher component. This component is used
  to launch an application from within a Delphi application. To actually
  run the desired program, the ShellExecuteEx function is used.

  Developing Custom Delphi 3 Components
  Copyright  1995-1997 by Ray Konopka
=========================================================================}

{$I DCDC.INC}

unit RkLaunch;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls, RkCommon;

type
  ELaunchError = class( Exception )              // Custom exception class
    ErrorCode : Integer;
  end;

  TShowMode = ( smNormal, smMaximized, smMinimized, smHide );

const
  ShowWindowModes : array[ TShowMode ] of Integer =
    ( sw_Normal, sw_ShowMaximized, sw_ShowMinimized, sw_Hide );

type
  TRkLauncher = class(TComponent)
  private
    FAboutInfo : TRkAboutInfo;
    FHInstance : THandle;
    FAction : string;
    FFileName : string;
    FParameters : string;
    FShowMode : TShowMode;
    FStartDir : string;
    FExitCode : DWord;
    FHProcess : THandle;
    FTimeout : Integer;
    FOnFinished : TNotifyEvent;
    FOnTimeout : TNotifyEvent;
  protected
    procedure Finished; dynamic;                 // Event dispatch methods
    procedure DoTimeout; dynamic;
  public
    constructor Create( AOwner : TComponent ); override;
    procedure Launch;

    // Read-only properties

    property HInstance : THandle
      read FHInstance;

    property ExitCode : DWord
      read FExitCode;

    property HProcess : THandle
      read FHProcess;
  published
    property About : TRkAboutInfo
      read FAboutInfo
      write FAboutInfo
      stored False;

    property Action : string
      read FAction
      write FAction;

    property FileName : string
      read FFileName
      write FFileName;

    property Parameters : string
      read FParameters
      write FParameters;

    property ShowMode : TShowMode
      read FShowMode
      write FShowMode
      default smNormal;

    property StartDir : string
      read FStartDir
      write FStartDir;

    property Timeout : Integer
      read FTimeout
      write FTimeout;

    property OnFinished : TNotifyEvent
      read FOnFinished
      write FOnFinished;

    property OnTimeout: TNotifyEvent
      read FOnTimeout
      write FOnTimeout;
  end;


implementation

{$IFDEF CHAPTER_11}
uses
  ShellApi;

function CreateLaunchError( ErrCode : Integer ) : ELaunchError;
var
  ErrMsg : string;
begin
  case ErrCode of
    0, 8 : ErrMsg := 'Out of Memory';
    2 : ErrMsg := 'File Not Found';
    3 : ErrMsg := 'Path Not Found';
    5 : ErrMsg := 'Sharing Violation';
    6 : ErrMsg := 'Library/Segment Error';
    10 : ErrMsg := 'Incorrect Version of Windows';
    11 : ErrMsg := 'Invalid File';
    12, 13 : ErrMsg := 'Invalid DOS Version';
    14, 31 : ErrMsg := 'File Type has No Association';
    15 : ErrMsg := 'Cannot Load a Real-Mode Application';
    16 : ErrMsg := 'Invalid Attempt to Launch Second Instance';
    19 : ErrMsg := 'Cannot Load a Compressed Executable File';
    20 : ErrMsg := 'Invalid Dynamic Link Library';
    21 : ErrMsg := 'Application Requires Win32';
    else
      ErrMsg := 'Unknown Error';
  end;
  Result := ELaunchError.Create( ErrMsg );
  Result.ErrorCode := ErrCode;
end;
{$ELSE}
uses
  ShellApi, LnchMsgs;


function CreateLaunchError( ErrCode : Integer ) : ELaunchError;
begin
  Result := ELaunchError.Create( LoadStr( SLaunchOutOfMemory + ErrCode ));
  Result.ErrorCode := ErrCode;
end;
{$ENDIF}

{=========================}
{== TRkLauncher Methods ==}
{=========================}

constructor TRkLauncher.Create( AOwner : TComponent );
begin
  inherited Create( AOwner );
  FShowMode := smNormal;
  FHInstance := 0;
  FAction := 'Open';                         // Open, or execute, the file
  FTimeout := Infinite;    // Do not stop waiting for process to terminate
  FHProcess := 0;
  FExitCode := 0;
end;


procedure TRkLauncher.Finished;
begin
  if Assigned( FOnFinished ) then
    FOnFinished( Self );
end;


procedure TRkLauncher.DoTimeout;
begin
  if Assigned( FOnTimeout ) then
    FOnTimeout( Self );
end;


procedure TRkLauncher.Launch;
var
  ShellInfo : TShellExecuteInfo;
begin
  FHInstance := 0;
  FHProcess := 0;
  FExitCode := 0;

  // Initialize the ShellInfo record

  FillChar( ShellInfo, SizeOf( TShellExecuteInfo ), 0 );
  ShellInfo.cbSize := SizeOf( TShellExecuteInfo );
  ShellInfo.fMask := SEE_MASK_NOCLOSEPROCESS;
  ShellInfo.Wnd := HWnd_Desktop;
  ShellInfo.lpVerb := PChar( FAction );
  ShellInfo.lpFile := PChar( FFileName );
  ShellInfo.lpParameters := PChar( FParameters );
  ShellInfo.lpDirectory := PChar( FStartDir );
  ShellInfo.nShow := ShowWindowModes[ FShowMode ];

  if ShellExecuteEx( @ShellInfo ) then
  begin
    FHInstance := ShellInfo.hInstApp;
    FHProcess := ShellInfo.hProcess;
    case WaitForSingleObject( FHProcess, FTimeout ) of
      wait_Failed:
        CreateLaunchError( GetLastError );

      wait_Object_0:
      begin
        GetExitCodeProcess( FHProcess, FExitCode );
        Finished;
      end;

      wait_Timeout:
        DoTimeout;
    end; { case }
  end
  else
    raise CreateLaunchError( ShellInfo.hInstApp );
end; {= TRkLauncher.Launch =}


end.
