{=========================================================================
  RkLblEdt Unit

  This unit implements a component editor for the RkLabel component. This
  unit defines the TRkLabelEditor class and the TRkLabelEditDlg class.
  The TRkLabelEditor displays the TRkLabelEditDlg form when the component
  is double-clicked or the first context menu item is chosen. The form
  provides a more visual way of editing the properties of the RkLabel
  component with immediate visual feedback.

  Developing Custom Delphi 3 Components
  Copyright  1995-1997 by Ray Konopka
=========================================================================}

unit RkLblEdt;

interface

uses
  SysUtils, Windows, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, RkLabel, ExtCtrls, Buttons, RkTrkBar, DsgnIntf;

type
  TRkLabelEditor = class( TComponentEditor )
    function GetVerbCount : Integer; override;
    function GetVerb( Index : Integer ) : string; override;
    procedure ExecuteVerb( Index : Integer ); override;
  end;


  TRkLabelEditDlg = class( TForm )
    GrpPreview : TGroupBox;
    GrpTextStyle : TRadioGroup;
    EdtCaption : TEdit;
    LblPreview : TRkLabel;
    GrpFontStyle : TGroupBox;
    ChkBold : TCheckBox;
    ChkItalic : TCheckBox;
    ChkStrikeout : TCheckBox;
    ChkUnderline : TCheckBox;
    Label1 : TLabel;
    GrpShadow : TGroupBox;
    TrkShadow : TRkTrackBar;
    CbxShadowColor : TComboBox;
    Label4 : TLabel;
    Label5 : TLabel;
    BtnOK : TButton;
    BtnCancel : TButton;
    GrpFontFace : TGroupBox;
    TrkPointSize : TRkTrackBar;
    CbxFonts : TComboBox;
    CbxFontColor : TComboBox;
    Label2 : TLabel;
    Label3 : TLabel;
    Label6 : TLabel;
    procedure EdtCaptionChange( Sender : TObject );
    procedure GrpTextStyleClick( Sender : TObject );
    procedure TrkPointSizeDrawTick( TrackBar : TRkTrackBar;
                                    Canvas : TCanvas;
                                    Location : TPoint; Index : Integer );
    procedure TrkPointSizeChange( Sender : TObject );
    procedure TrkShadowChange( Sender : TObject );
    procedure FormCreate( Sender : TObject );
    procedure CbxFontsChange( Sender : TObject );
    procedure ChkBoldClick( Sender : TObject );
    procedure ChkItalicClick( Sender : TObject );
    procedure ChkStrikeoutClick( Sender : TObject );
    procedure CbxColorDrawItem( Control : TWinControl; Index : Integer;
                                Rect : TRect; State : TOwnerDrawState );
    procedure ChkUnderlineClick( Sender : TObject );
    procedure CbxFontColorChange( Sender : TObject );
    procedure CbxShadowColorChange( Sender : TObject );
  public
    procedure InitSettings;
  end;


procedure Register;

implementation

{$R *.DFM}


{============================}
{== TRkLabelEditor Methods ==}
{============================}

function TRkLabelEditor.GetVerbCount : Integer;
begin
  Result := 4;           // Return the number of new menu items to display
end;


function TRkLabelEditor.GetVerb( Index : Integer ) : string;
begin
  case Index of
    0: Result := 'Edit &Label...';
    1: Result := 'Ra&ised';
    2: Result := '&Recessed';
    3: Result := 'Shado&w';
  end;
end;


{=========================================================================
  TRkLabelEditor.ExecuteVerb

  This method will be invoked by Delphi whenever the user selects one of
  the newly created item on the components context menu while in the
  Form Designer. If the first item is selected (i.e. Edit Label...), then
  the TRkLabelEditDlg form is displayed. If one of the other menu items
  is selected, the label's TextStyle property is changed appropriately.
=========================================================================}

procedure TRkLabelEditor.ExecuteVerb( Index : Integer );
var
  D : TRkLabelEditDlg;

  procedure CopyLabel( Dest, Source : TRkLabel );
  begin
    Dest.Caption := Source.Caption;
    Dest.Font := Source.Font;
    Dest.TextStyle := Source.TextStyle;
    Dest.ShadowColor := Source.ShadowColor;
    Dest.ShadowDepth := Source.ShadowDepth;
  end;

begin
  case Index of
    0:                                    // User selected "Edit Label..."
    begin
      D := TRkLabelEditDlg.Create( Application );

      try
        // Copy component attributes to the LblPreview component
        CopyLabel( D.LblPreview, Component as TRkLabel );

        // Set the dialog's Caption to reflect component being edited
        D.Caption := Component.Owner.Name +'.'+ Component.Name +D.Caption;

        D.InitSettings;               // Update all controls on dialog box

        if D.ShowModal = mrOK then               // Display the dialog box
        begin
          CopyLabel( Component as TRkLabel, D.LblPreview );
          // Tell the Form Designer to set the Modified flag for the form
          Designer.Modified;
        end;
      finally
        D.Free;
      end;
    end;

    1:                                           // User selected "Raised"
    begin
      TRkLabel( Component ).TextStyle := tsRaised;
      Designer.Modified;
    end;

    2:                                         // User selected "Recessed"
    begin
      TRkLabel( Component ).TextStyle := tsRecessed;
      Designer.Modified;
    end;

    3:                                           // User selected "Shadow"
    begin
      TRkLabel( Component ).TextStyle := tsShadow;
      Designer.Modified;
    end;
  end; { case }
end; {= TRkLabelEditor.ExecuteVerb =}



{=================================================}
{== Implementation Specific Types and Constants ==}
{=================================================}

const            // Point size strings to display as ticks along track bar
  PointSizes : array[ 0..17 ] of string[ 2 ] =
    ( '6', '8', '9', '10', '11', '12', '14', '16', '18', '20',
      '22', '24', '28', '32', '40', '48', '64', '72' );

type                              // Supports owner draw color combo boxes
  TColorRec = record
    Color : TColor;
    Name : string[ 7 ];
  end;

const
  Colors : array[ 0..15 ] of TColorRec = (
    ( Color : clBlack;   Name : 'Black' ),
    ( Color : clMaroon;  Name : 'Maroon' ),
    ( Color : clGreen;   Name : 'Green' ),
    ( Color : clOlive;   Name : 'Olive' ),
    ( Color : clNavy;    Name : 'Navy' ),
    ( Color : clPurple;  Name : 'Purple' ),
    ( Color : clTeal;    Name : 'Teal' ),
    ( Color : clGray;    Name : 'Gray' ),
    ( Color : clSilver;  Name : 'Silver' ),
    ( Color : clRed;     Name : 'Red' ),
    ( Color : clLime;    Name : 'Lime' ),
    ( Color : clYellow;  Name : 'Yellow' ),
    ( Color : clBlue;    Name : 'Blue' ),
    ( Color : clFuchsia; Name : 'Fuchsia' ),
    ( Color : clAqua;    Name : 'Aqua' ),
    ( Color : clWhite;   Name : 'White' ) );



{=============================}
{== TRkLabelEditDlg Methods ==}
{=============================}

{=========================================================================
  NOTE:
  All changes made through the control on this dialog box affect only
  the preview label (LblPreview).  Only if the OK button is pressed
  are the changes reflected in the selected component.
=========================================================================}

procedure TRkLabelEditDlg.FormCreate( Sender : TObject );
begin
  CbxFonts.Items := Screen.Fonts;
end;


procedure TRkLabelEditDlg.InitSettings;

  function PositionFromPointSize( P : Integer ) : Integer;
  var
    I : Integer;
  begin
    I := 0;
    while ( I < 18 ) and ( StrToInt( PointSizes[ I ] ) <> P ) do
      Inc( I );
    if I = 18 then
      Result := 2
    else
      Result := I;
  end;


  function IndexFromColor( C : TColor ) : Integer;
  var
    I : Integer;
  begin
    I := 0;
    while (I < 16) and ( ColorToRGB(Colors[I].Color) <> ColorToRGB(C) ) do
      Inc( I );
    if I = 16 then
      Result := 0
    else
      Result := I;
  end;

begin {= TRkLabelEditDlg.InitSettings =}
  EdtCaption.Text := LblPreview.Caption;
  CbxFonts.ItemIndex := CbxFonts.Items.IndexOf( LblPreview.Font.Name );
  CbxFontColor.ItemIndex := IndexFromColor( LblPreview.Font.Color );
  TrkPointSize.Position := PositionFromPointSize( LblPreview.Font.Size );

  // Font Styles
  ChkBold.Checked := fsBold in LblPreview.Font.Style;
  ChkItalic.Checked := fsItalic in LblPreview.Font.Style;
  ChkStrikeout.Checked := fsStrikeout in LblPreview.Font.Style;
  ChkUnderline.Checked := fsUnderline in LblPreview.Font.Style;

  // Text Style
  GrpTextStyle.ItemIndex := Ord( LblPreview.TextStyle );

  // Shadow Options
  CbxShadowColor.ItemIndex := IndexFromColor( LblPreview.ShadowColor );
  TrkShadow.Position := LblPreview.ShadowDepth;
end; {= TRkLabelEditDlg.InitSettings =}


procedure TRkLabelEditDlg.EdtCaptionChange( Sender : TObject );
begin
  LblPreview.Caption := EdtCaption.Text;
end;


procedure TRkLabelEditDlg.CbxFontsChange( Sender : TObject );
begin
  LblPreview.Font.Name := CbxFonts.Text;
end;


{=========================================================================
  TRkLabelEditDlg.TrkPointSizeDrawTick

  Owner draw method used to display point size values at each tick mark.
=========================================================================}

procedure TRkLabelEditDlg.TrkPointSizeDrawTick( TrackBar : TRkTrackBar;
  Canvas : TCanvas; Location : TPoint; Index : Integer );
var
  W : Integer;
begin
  Canvas.Brush.Color := TrackBar.Color;
  Canvas.Font.Name := 'Small Fonts';
  Canvas.Font.Size := 7;
  Canvas.Font.Style := [];
  W := Canvas.TextWidth( PointSizes[ Index ] );
  Canvas.TextOut( Location.X - (W div 2), 1, PointSizes[ Index ] );
end;


procedure TRkLabelEditDlg.TrkPointSizeChange( Sender : TObject );
begin
  LblPreview.Font.Size := StrToInt( PointSizes[ TrkPointSize.Position ] );
end;


procedure TRkLabelEditDlg.ChkBoldClick( Sender : TObject );
begin
  if ChkBold.Checked then
    LblPreview.Font.Style := LblPreview.Font.Style + [ fsBold ]
  else
    LblPreview.Font.Style := LblPreview.Font.Style - [ fsBold ]
end;


procedure TRkLabelEditDlg.ChkItalicClick( Sender : TObject );
begin
  if ChkItalic.Checked then
    LblPreview.Font.Style := LblPreview.Font.Style + [ fsItalic ]
  else
    LblPreview.Font.Style := LblPreview.Font.Style - [ fsItalic ]
end;


procedure TRkLabelEditDlg.ChkStrikeoutClick( Sender : TObject );
begin
  if ChkStrikeout.Checked then
    LblPreview.Font.Style := LblPreview.Font.Style + [ fsStrikeout ]
  else
    LblPreview.Font.Style := LblPreview.Font.Style - [ fsStrikeout ]
end;


procedure TRkLabelEditDlg.ChkUnderlineClick( Sender : TObject );
begin
  if ChkUnderline.Checked then
    LblPreview.Font.Style := LblPreview.Font.Style + [ fsUnderline ]
  else
    LblPreview.Font.Style := LblPreview.Font.Style - [ fsUnderline ]
end;


procedure TRkLabelEditDlg.GrpTextStyleClick( Sender : TObject );
begin
  LblPreview.TextStyle := TTextStyle( GrpTextStyle.ItemIndex );

  TrkShadow.Enabled := LblPreview.TextStyle = tsShadow;
  CbxShadowColor.Enabled := LblPreview.TextStyle = tsShadow;
end;


procedure TRkLabelEditDlg.TrkShadowChange( Sender : TObject );
begin
  LblPreview.ShadowDepth := TrkShadow.Position;
end;


{=========================================================================
  TRkLabelEditDlg.CbxColorDrawItem

  This owner-draw method is used to draw color entries for both the Font
  Color and Shadow Color combo boxes.
=========================================================================}

procedure TRkLabelEditDlg.CbxColorDrawItem( Control : TWinControl;
                                            Index : Integer; Rect : TRect;
                                            State : TOwnerDrawState );
var
  R : TRect;
begin
  R := Rect;
  InflateRect( R, -2, -2 );
  R.Right := 20;
  with ( Control as TComboBox ).Canvas do
  begin
    FillRect( Rect );                         // Clear list item rectangle
    Brush.Color := Colors[ Index ].Color;          // Select correct color
    Rectangle( R.Left, R.Top, R.Right, R.Bottom ); // Draw color rectangle
    if odSelected in State then
      Brush.Color := clHighlight
    else
      Brush.Color := (Control as TComboBox).Color;
    TextOut( Rect.Left + 24, Rect.Top + 2, Colors[ Index ].Name );
  end;
end;


procedure TRkLabelEditDlg.CbxFontColorChange( Sender : TObject );
begin
  LblPreview.Font.Color := Colors[ CbxFontColor.ItemIndex ].Color;
end;


procedure TRkLabelEditDlg.CbxShadowColorChange( Sender : TObject );
begin
  LblPreview.ShadowColor := Colors[ CbxShadowColor.ItemIndex ].Color;
end;



{========================}
{== Register Procedure ==}
{========================}

procedure Register;
begin
  RegisterComponentEditor( TRkLabel, TRkLabelEditor );
end;


end.
