{=========================================================================
  RkROGrid Unit

  This unit implements the TRkDBReadOnlyGrid component, which, as its
  name suggests, is a read-only data-aware grid component.  When this
  component is connected to a dataset (via the DataSource property), the
  entire contents of the dataset is displayed in a familiar grid-like
  fashion.

  NOTE: While this component is by no means a replacement for the
        standard DBGrid, it does illustrate how the TDataLink class
        is used to access the entire contents of a dataset.

  Developing Custom Delphi 3 Components
  Copyright  1995-1997 by Ray Konopka
=========================================================================}

{$I DCDC.INC}

unit RkROGrid;

interface

uses
  Windows, Messages, Classes, Graphics, Controls, ExtCtrls, RkCommon, DB;

type
  TRkDBReadOnlyGrid = class;                    // Forward class reference

  TRkROGridDataLink = class( TDataLink )
  private
    FGrid : TRkDBReadOnlyGrid;
    function GetFieldCount : Integer;
    function GetFields( I : Integer ) : TField;
  protected
    procedure ActiveChanged; override;
    procedure DatasetChanged; override;
    procedure DatasetScrolled( Distance : Integer ); override;
    procedure LayoutChanged; override;
    procedure RecordChanged( Field : TField ); override;
  public
    constructor Create( AGrid : TRkDBReadOnlyGrid );

    property FieldCount : Integer
      read GetFieldCount;

    property Fields[ I : Integer ] : TField
      read GetFields;
  end;


  TRkDBReadOnlyGrid = class( TCustomPanel )
  private
    FAboutInfo : TRkAboutInfo;
    FDataLink : TRkROGridDataLink;
    FRowHeight : Integer;
    FColStart : Integer;
    FBitmap : TBitmap;
    function GetDataSource : TDataSource;
    procedure SetDataSource( Value : TDataSource );

    procedure UpdateRowHeight;
    procedure UpdateBufferCount;
    procedure UpdateVertScrollBar;
    procedure UpdateHorzScrollBar;
    procedure DataChanged;

    procedure WMVScroll( var Msg : TWMVScroll ); message wm_VScroll;
    procedure WMHScroll( var Msg : TWMHScroll ); message wm_HScroll;
    procedure CMFontChanged( var Msg : TMessage ); message cm_FontChanged;
  protected
    procedure CreateParams( var Params : TCreateParams ); override;
    procedure CreateWnd; override;
    procedure Paint; override;
    function VisibleRows : Integer;
    procedure Resize; override;
  public
    constructor Create( AOwner : TComponent ); override;
    destructor Destroy; override;
  published
    property About : TRkAboutInfo
      read FAboutInfo
      write FAboutInfo
      stored False;

    property DataSource : TDataSource
      read GetDataSource
      write SetDataSource;

    // Inherited properties and events
    property Align;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Color;
    property Font;
    property Height default 110;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDrag;
  end;


implementation

uses
  Forms;

{===============================}
{== TRkROGridDataLink Methods ==}
{===============================}

constructor TRkROGridDataLink.Create( AGrid : TRkDBReadOnlyGrid );
begin
  inherited Create;
  FGrid := AGrid;           // Store the grid control using this data link
end;


function TRkROGridDataLink.GetFieldCount : Integer;
begin
  Result := Dataset.FieldCount;
end;


function TRkROGridDataLink.GetFields( I : Integer ) : TField;
begin
  if ( I >= 0 ) and ( I < Dataset.FieldCount ) then
    Result := DataSet.Fields[ I ]
  else
    Result := nil;
end;


{=========================================================================
  TRkROGridDataLink.ActiveChanged

  This method is called whenever the Active property of the associated
  dataset is changed.

  In this instance, both of the grid's scroll bars need to be updated.
  The grid also needs to be repainted to reflect the contents of the
  new dataset.
=========================================================================}

procedure TRkROGridDataLink.ActiveChanged;
begin
  FGrid.UpdateVertScrollBar;
  FGrid.UpdateHorzScrollBar;
  FGrid.Invalidate;
end;


{=========================================================================
  TRkROGridDataLink.DatasetChanged

  This method is called when the dataset changes as a result of a record
  being inserted or deleted. This method also gets called when a record
  is posted to the dataset. In this case, this method is called after
  the RecordChanged method.

  In this instance, the grid's DataChange method is called to update the
  buffer count, update the vertical scroll bar, and repaint the grid.
  Repainting is necessary to update the highlight bar.
=========================================================================}

procedure TRkROGridDataLink.DatasetChanged;
begin
  FGrid.DataChanged;
end;


{=========================================================================
  TRkROGridDataLink.DatasetChanged

  This method is called whenever the dataset is moved to a new record.

  In this instance, the grid's DataChange method is called to update the
  buffer count, update the vertical scroll bar, and repaint the grid.
  Repainting is necessary to update the highlight bar.
=========================================================================}

procedure TRkROGridDataLink.DataSetScrolled( Distance : Integer );
begin
  FGrid.DataChanged;
end;


{=========================================================================
  TRkROGridDataLink.LayoutChanged

  This method is called whenever a change is made to a property in one of
  the dataset's field objects that causes the appearance of the field's
  data to change. For example, if the Visible property of a field object
  is set to False, the corresponding column should be removed from the
  grid's display.

  In this instance, the grid is repainted to reflect the change.
=========================================================================}

procedure TRkROGridDataLink.LayoutChanged;
begin
  FGrid.Invalidate;
end;


{=========================================================================
  TRkROGridDataLink.RecordChanged

  This method is called whenever a field in the current record is
  modified. This method is also called when the current record in the
  dataset is posted to the database. In this case, this method is
  called before the DatasetChanged method.

  In this instance, the grid is repainted to reflect any changes made
  to the current record.
=========================================================================}

procedure TRkROGridDataLink.RecordChanged( Field : TField );
begin
  FGrid.Invalidate;
end;



{===============================}
{== TRkDBReadOnlyGrid Methods ==}
{===============================}

constructor TRkDBReadOnlyGrid.Create( AOwner : TComponent );
begin
  inherited Create( AOwner );
  // No caption needed and should not be placed in a DBCtrlGrid
  ControlStyle := ControlStyle - [ csSetCaption, csReplicatable ];

  Height := 110;
  BorderStyle := bsSingle;
  BevelOuter := bvNone;
  TabStop := False;
  FColStart := 0;

  FBitmap := TBitmap.Create;    // Use off-screen bitmap to reduce flicker
  UpdateRowHeight;              // Calculate row height using current font

  FDataLink := TRkROGridDataLink.Create( Self );   // Create the data link
end;


procedure TRkDBReadOnlyGrid.CreateParams( var Params : TCreateParams );
begin
  inherited CreateParams( Params );
  // Instruct Windows to display a vertical and horizontal scroll bar
  Params.Style := Params.Style or ws_VScroll or ws_HScroll;
end;


procedure TRkDBReadOnlyGrid.CreateWnd;
begin
  inherited CreateWnd;
  UpdateBufferCount;
  UpdateVertScrollBar;
  UpdateHorzScrollBar;
end;


destructor TRkDBReadOnlyGrid.Destroy;
begin
  FBitmap.Free;
  FDataLink.Free;                 // Don't forget to release the data link
  FDataLink := nil;
  inherited Destroy;
end;


function TRkDBReadOnlyGrid.GetDataSource : TDataSource;
begin
  Result := FDataLink.DataSource;
end;


procedure TRkDBReadOnlyGrid.SetDataSource( Value : TDataSource );
begin
  if FDatalink.DataSource <> Value then
  begin
    FDataLink.DataSource := Value;

    if Value <> nil then
      Value.FreeNotification( Self );
  end;
end;


procedure TRkDBReadOnlyGrid.DataChanged;
begin
  if not HandleAllocated then
    Exit;

  // The current record has been changed, adjust the vertical scroll bar.

  UpdateVertScrollBar;
  Invalidate;
end;


procedure TRkDBReadOnlyGrid.UpdateBufferCount;
begin
  // Request that the data link's internal buffer be set to the
  // number of rows currently visible in the grid.
  FDataLink.BufferCount := VisibleRows;
end;


procedure TRkDBReadOnlyGrid.Paint;
var
  Row, Col, I, X : Integer;
  TextXPos, SaveActive : Integer;
  StartRow, StopRow : Integer;
  LineWidth, LineHeight : Integer;
  BtnRect, CellRect : TRect;


  function GetFieldWidth( Index : Integer ) : Integer;
  var
    TM : TTextMetric;
  begin
    // Determines a column's width using the current font and the
    // corresponding field's DisplayWidth property

    GetTextMetrics( Canvas.Handle, TM );
    Result := FDataLink.Fields[ Index ].DisplayWidth *
              ( Canvas.TextWidth( '0' ) - TM.tmOverhang ) +
              TM.tmOverhang + 4;
  end;


  function Min( A, B : Integer ) : Integer;
  begin
    if A < B then
      Result := A
    else
      Result := B;
  end;


  procedure DrawHeaderCell( R : TRect );
  begin
    // Draws a cell representing a column header.
    with FBitmap.Canvas, R do
    begin
      Brush.Color := clBtnFace;
      FillRect( R );

      Pen.Color := clBtnHighlight;
      MoveTo( Left, Bottom - 1 );
      LineTo( Left, Top );
      LineTo( Right, Top );

      Pen.Color := clBtnText;
      MoveTo( Right - 1, Top  );
      LineTo( Right - 1, Bottom - 1 );
      LineTo( Left - 1, Bottom - 1 );

      Pen.Color := clBtnFace;
      MoveTo( Left + 1, Bottom - 2 );
      LineTo( Left + 1, Top + 1 );
      LineTo( Right - 1, Top + 1 );

      Pen.Color := clBtnShadow;
      MoveTo( Right - 2, Top + 1 );
      LineTo( Right - 2, Bottom - 2 );
      LineTo( Left, Bottom - 2 );
    end;
  end; {= DrawHeaderCell =}


  procedure DrawVertDottedLine( X, Y1, Y2 : Integer );
  var
    Y : Integer;
  begin
    // Draws a *true* dotted vertical line
    for Y := Y1 to Y2 do
      if Y mod 2 <> 0 then
        FBitmap.Canvas.Pixels[ X, Y ] := clBtnShadow;
  end;


  procedure DrawHorzDottedLine( Y, X1, X2 : Integer );
  var
    X : Integer;
  begin
    // Draws a *true* dotted horizontal line
    for X := X1 to X2 do
      if X mod 2 <> 0 then
        FBitmap.Canvas.Pixels[ X, Y ] := clBtnShadow;
  end;


begin {= TRkDBReadOnlyGrid.Paint =}
  // Draw grid cells on an offscreen bitmap to eliminate flicker
  FBitmap.Width := ClientWidth;
  FBitmap.Height := ClientHeight;
  FBitmap.Canvas.Pen.Style := psClear;
  FBitmap.Canvas.Brush.Color := Color;
  FBitmap.Canvas.Rectangle( 0, 0, FBitmap.Width + 1, FBitmap.Height + 1 );

  if ( FDataLink = nil ) or not FDataLink.Active then
  begin
    // Grid is not connect to a dataset, display single empty header cell
    FBitmap.Canvas.Pen.Style := psSolid;
    BtnRect := Rect( 0, 0, Width div 4, FRowHeight );
    DrawHeaderCell( BtnRect );
    DrawHorzDottedLine( 2 * FRowHeight, 0, Width div 4 - 1);
    DrawVertDottedLine( Width div 4 - 1, FRowHeight, 2 * FRowHeight );
    Canvas.Draw( 0, 0, FBitmap );
    Exit;                                          // Exit Paint procedure
  end;

  with FBitmap.Canvas do
  begin
    Font := Self.Font;
    Brush.Color := Self.Color;

    // Display column headings

    Pen.Style := psSolid;
    X := 0;
    for Col := FColStart to FDataLink.FieldCount - 1 do
    begin
      if FDataLink.Fields[ Col ].Visible then
      begin
        BtnRect := Rect( X, 0, X + GetFieldWidth( Col ), FRowHeight );
        DrawHeaderCell( BtnRect );
        InflateRect( BtnRect, -2, -2 );
        TextRect( BtnRect, BtnRect.Left, 2,
                  FDataLink.Fields[ Col ].DisplayLabel );
        X := X + GetFieldWidth( Col );
      end;
    end;
    Pen.Style := psClear;

    // Determine which rows to display

    StartRow := FDataLink.ActiveRecord;   // Start with the current record
    for I := 1 to VisibleRows do
      if StartRow > 0 then
        Dec( StartRow );                   // Move back as far as possible
    // Set StopRow to fill up each visible row
    StopRow := Min( FDataLink.RecordCount, StartRow + VisibleRows - 1 );


    // Iterate through each row to be displayed in the grid

    for Row := StartRow to StopRow do
    begin
      X := 0;

      // Save the current record number
      SaveActive := FDataLink.ActiveRecord;
      try
        // Point the *data link* to another row in the dataset. This gives
        // us access to the data stored in the other rows without changing
        // the current record in the *dataset*.

        FDataLink.ActiveRecord := Row;

        if FDataLink.ActiveRecord = SaveActive then
        begin
          // Highlight the dataset's current record
          Brush.Color := clHighlight;
          Font.Color := clHighlightText;
        end
        else
        begin
          Brush.Color := Self.Color;
          Font := Self.Font;
        end;

        // Iterate through each column (i.e. field) in the dataset

        for Col := FColStart to FDataLink.FieldCount - 1 do
        begin
          // Don't display the column if the Visible property is False
          if FDataLink.Fields[ Col ].Visible then
          begin
            // Calculate dimensions of cell
            CellRect := Rect( X, ( Row + 1 ) * FRowHeight,
                              X + GetFieldWidth( Col ),
                              ( Row + 2 ) * FRowHeight - 1 );

            // Determine text alignment from field's Alignment property
            TextXPos := 0;
            case FDataLink.Fields[ Col ].Alignment of
              taLeftJustify:
              begin
                SetTextAlign( Handle, ta_Left );
                TextXPos := CellRect.Left + 2;
              end;

              taRightJustify:
              begin
                SetTextAlign( Handle, ta_Right );
                TextXPos := CellRect.Right - 2;
              end;

              taCenter:
              begin
                SetTextAlign( Handle, ta_Center );
                TextXPos := ( CellRect.Right - CellRect.Left ) div 2;
              end;
            end; { case }

            // Display the contents of the cell using the corresponding
            // field object's DisplayText property.

            TextRect( CellRect, TextXPos, CellRect.Top + 2,
                      FDataLink.Fields[ Col ].DisplayText );
            X := X + GetFieldWidth( Col );
          end;
        end;
      finally
        // Restore the data link's active record
        FDataLink.ActiveRecord := SaveActive;
      end;
    end; { for R }


    // Draw vertical grid lines

    X := 0;
    LineHeight := ( Min( VisibleRows, FDataLink.RecordCount ) + 1 ) *
                  FRowHeight - 1;
    for Col := FColStart to FDataLink.FieldCount - 1 do
    begin
      if FDataLink.Fields[ Col ].Visible then
      begin
        X := X + GetFieldWidth( Col );
        DrawVertDottedLine( X - 1, FRowHeight, LineHeight );
      end;
    end;

    // Draw horizontal grid lines

    LineWidth := X - 1;
    for Row := 0 to Min( VisibleRows, FDataLink.RecordCount ) - 1 do
      DrawHorzDottedLine( ( Row + 2 ) * FRowHeight - 1, 0, LineWidth );

  end; { with Canvas }

  Canvas.Draw( 0, 0, FBitmap );  // Draw offscreen bitmap into client area
  
end; {= TRkDBReadOnlyGrid.Paint =}


procedure TRkDBReadOnlyGrid.UpdateVertScrollBar;
var
  Pos : Integer;
begin
  if FDatalink.Active and HandleAllocated then
  begin
    SetScrollRange( Self.Handle, sb_Vert, 0, 4, False );

    if FDataLink.Dataset.BOF then
      Pos := 0
    else if FDataLink.Dataset.EOF then
      Pos := 4
    else
      Pos := 2;

    if GetScrollPos( Self.Handle, sb_Vert ) <> Pos then
      SetScrollPos( Self.Handle, sb_Vert, Pos, True );
  end;
end;


procedure TRkDBReadOnlyGrid.UpdateHorzScrollBar;
begin
  if FDatalink.Active and HandleAllocated then
  begin
    SetScrollRange( Self.Handle, sb_Horz, 0, FDataLink.FieldCount - 1,
                    False );

    FColStart := 0;
    if GetScrollPos( Self.Handle, sb_Horz ) <> FColStart then
    begin
      SetScrollPos( Self.Handle, sb_Horz, FColStart, True );
      Invalidate;
    end;
  end;
end;


procedure TRkDBReadOnlyGrid.WMVScroll( var Msg : TWMVScroll );
begin
  if FDatalink.Active then
  begin
    case Msg.ScrollCode of
      sb_LineUp:
        FDataLink.DataSet.MoveBy( -FDataLink.ActiveRecord - 1 );

      sb_LineDown:
        FDataLink.DataSet.MoveBy( FDataLink.RecordCount -
                                  FDataLink.ActiveRecord );

      sb_PageUp:
        FDataLink.DataSet.MoveBy( -VisibleRows );

      sb_PageDown:
        FDataLink.DataSet.MoveBy( VisibleRows );

      sb_ThumbPosition:
      begin
        case Msg.Pos of
          0: FDataLink.DataSet.First;
          1: FDataLink.DataSet.MoveBy( -VisibleRows );
          2: Exit;
          3: FDataLink.DataSet.MoveBy( VisibleRows );
          4: FDataLink.DataSet.Last;
        end;
      end;

      sb_Bottom:
        FDataLink.DataSet.Last;

      sb_Top :
        FDataLink.DataSet.First;
    end;
  end;
end;


procedure TRkDBReadOnlyGrid.WMHScroll( var Msg : TWMHScroll );
begin
  if FDatalink.Active then
  begin
    case Msg.ScrollCode of
      sb_LineLeft, sb_PageLeft:
      begin
        Dec( FColStart );
        if FColStart < 0 then
          FColStart := 0;
      end;

      sb_LineRight, sb_PageRight:
      begin
        Inc( FColStart );
        if FColStart > FDataLink.FieldCount - 1 then
          FColStart := FDataLink.FieldCount - 1;
      end;

      sb_ThumbPosition:
        FColStart := Msg.Pos;

      sb_Bottom:
        FColStart := FDataLink.FieldCount - 1;

      sb_Top :
        FColStart := 0;
    end;

    if GetScrollPos( Self.Handle, sb_Horz ) <> FColStart then
    begin
      SetScrollPos( Self.Handle, sb_Horz, FColStart, True );
      Invalidate;
    end;

  end;
end;


procedure TRkDBReadOnlyGrid.UpdateRowHeight;
var
  Canvas : TCanvas;
begin
  // Calculate the row height using the height of the current font plus
  // extra space to accommodate the grid lines.
  Canvas := TCanvas.Create;
  try
    Canvas.Handle := GetDC( 0 );
    try
      Canvas.Font := Font;
      FRowHeight := Canvas.TextHeight( 'Yy' ) + 5;
    finally
      ReleaseDC( 0, Canvas.Handle );
    end;
  finally
    Canvas.Free;
  end;
end;


procedure TRkDBReadOnlyGrid.CMFontChanged( var Msg : TMessage );
begin
  inherited;
  UpdateRowHeight;
end;


function TRkDBReadOnlyGrid.VisibleRows : Integer;
begin
  // Must take into account the header cells and the horizontal
  // scroll bar when determining the number of visible rows.
  Result := (Height - GetSystemMetrics(sm_CyHScroll)) div FRowHeight - 1;
end;


procedure TRkDBReadOnlyGrid.Resize;
begin
  inherited Resize;

  // Resizing the grid will cause the number of visible rows to change.
  // Therefore, the data link's buffer size needs to be adjusted.
  UpdateBufferCount;
end;


end.
